package bucket

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"regexp"
	"sort"
	"time"

	"github.com/go-redis/redis/v8"
	"github.com/open-io/oio-exporter/account"
	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("bucket")

var scriptLua = redis.NewScript(`
  local cursor = ARGV[1]
  local count = ARGV[2]
  local res = {}
  local n_buckets
  repeat
		local buckets = redis.call("zscan", "buckets:", cursor, "COUNT", count)
    cursor = tonumber(buckets[1])
    n_buckets = #(buckets[2])
    for i, bucket in ipairs(buckets[2]) do
		  if i % 2 == 1 then
        local ok = true -- by default, allow bucket

        -- KEYS are include patterns
        if #KEYS > 0 then
          ok = false -- by default, reject bucket when there are include patterns
          for i, r in ipairs(KEYS) do
            if bucket:match(r) then
              ok = true -- allow bucket
              break
            end
          end
        end

        -- ARGV are reject patterns
        if ok then
          for i = 3, #ARGV do
            if bucket:match(ARGV[i]) then
              ok = false -- reject bucket
              break
            end
          end
        end

        if ok then
					local values = redis.call("hmget", "bucket:" .. bucket, "account", "objects", "bytes")
					if #values == 3 and type(values[1]) == 'string' then
	          res[bucket] = {
		          account = values[1],
			        objects = tonumber(values[2]),
				      bytes = tonumber(values[3])
					  }
					end
        end
      end
		end
  until cursor == 0 or n_buckets > count/2

  return cjson.encode({cursor = cursor, buckets = res})
`)

var confTimeout = 30 * time.Second
var confInterval = 15 * time.Minute
var confIntervalScan = 2 * time.Second
var confRefreshSentinel = 5 * time.Minute
var confFiltersAllow common.ArrayFlags
var confFiltersReject common.ArrayFlags
var confScanCount = uint(1000)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "bucket.timeout", confTimeout, "[bucket] operation timeout")
	fs.DurationVar(&confInterval, "bucket.interval", confInterval, "[bucket] interval between polls")
	fs.DurationVar(&confIntervalScan, "bucket.interval.scan", confIntervalScan, "[bucket] limited interval between scan polls")
	fs.DurationVar(&confRefreshSentinel, "bucket.refresh", confRefreshSentinel, "[bucket] interval between refresh of sentinel state")
	fs.Var(&confFiltersAllow, "bucket.filter.allow", "[bucket] regex(es) to select which buckets to gather. Default no filtering")
	fs.Var(&confFiltersReject, "bucket.filter.reject", "[bucket] regex(es) to limit which buckets not to gather. Default no filtering")
	fs.UintVar(&confScanCount, "bucket.scan.count", confScanCount, "[bucket] Number of entries for scan() redis function to return")
}

var Factory = &bucketDaemonFactory{}

type bucketDaemonFactory struct{}

func (df *bucketDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})
	namespace := inventory.Namespace

	localRedis := []string{}
	for _, service := range inventory.Services["redis"] {
		localRedis = append(localRedis, fmt.Sprintf("%s:%d", service.IP, service.Port))
	}

	for _, service := range inventory.Services["redissentinel"] {
		params[service.ID] = &bucketDaemonParams{
			Namespace:   namespace,
			ServiceID:   service.ID,
			ServiceIP:   service.IP,
			ServicePort: service.Port,
			LocalRedis:  localRedis,
		}
	}

	return params
}

func (df *bucketDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*bucketDaemonParams)

	return &bucketDaemon{
		namespace:   params.Namespace,
		serviceID:   params.ServiceID,
		serviceIP:   params.ServiceIP,
		servicePort: params.ServicePort,
		LocalRedis:  params.LocalRedis,
	}
}

type bucketDaemonParams struct {
	Namespace   string
	ServiceID   string
	ServiceIP   string
	ServicePort int
	LocalRedis  []string
}

type bucketDaemon struct {
	daemon.BaseDaemon

	namespace   string
	serviceID   string
	serviceIP   string
	servicePort int
	serviceAddr string
	masterName  string
	slaves      []string
	lastUpdate  time.Time
	LocalRedis  []string

	redisClient    *redis.Client
	sentinelClient *redis.SentinelClient
	cursor         uint64

	filtersAllowAccount  []*regexp.Regexp
	filtersRejectAccount []*regexp.Regexp
}

func (d *bucketDaemon) Interval() time.Duration {
	if d.cursor == 0 {
		return confInterval
	}

	return confIntervalScan
}

func (d *bucketDaemon) Init(ctx context.Context) error {
	d.serviceAddr = fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)

	d.filtersAllowAccount = []*regexp.Regexp{}
	for _, r := range account.ConfFiltersAllow {
		re, err := regexp.Compile(r)
		if err != nil {
			log.Errorf("Unable to compile account Allow filter: %v", err)
			continue
		}
		d.filtersAllowAccount = append(d.filtersAllowAccount, re)
	}
	log.Debugf("FiltersAllowAccount: %+v", d.filtersAllowAccount)

	d.filtersRejectAccount = []*regexp.Regexp{}
	for _, r := range account.ConfFiltersReject {
		re, err := regexp.Compile(r)
		if err != nil {
			log.Errorf("Unable to compile account Allow filter: %v", err)
			continue
		}
		d.filtersRejectAccount = append(d.filtersRejectAccount, re)
	}
	log.Debugf("FiltersRejectAccount: %+v", d.filtersRejectAccount)

	d.sentinelClient = redis.NewSentinelClient(&redis.Options{
		Addr:         d.serviceAddr,
		DialTimeout:  confTimeout,
		ReadTimeout:  confTimeout,
		WriteTimeout: confTimeout,
	}).WithContext(ctx)

	return nil
}

func (d *bucketDaemon) updateSentinel(ctx context.Context) error {
	forceUpdate := false
	if d.lastUpdate.IsZero() || time.Since(d.lastUpdate) > confRefreshSentinel {
		forceUpdate = true
		log.Debugf("Force sentinel refresh")
	}

	if d.masterName == "" || forceUpdate {
		masters, err := d.sentinelClient.Masters(ctx).Result()
		if err != nil {
			return err
		}
		if len(masters) < 1 {
			return fmt.Errorf("Sentinel returned no master")
		}
		master := masters[0].([]interface{})
		if len(master) < 2 {
			return fmt.Errorf("Sentinel incomplete master informations")
		}
		d.masterName = master[1].(string)
		log.Debugf("got sentinel master name: %s", d.masterName)
	}

	if d.slaves == nil || forceUpdate {
		slaves, err := d.sentinelClient.Slaves(ctx, d.masterName).Result()
		if err != nil {
			return err
		}

		d.slaves = make([]string, len(slaves))
		for i, v := range slaves {
			slave := v.([]interface{})
			ipPort := slave[3].(string) + ":" + slave[5].(string)
			log.Debugf("slaves %s", ipPort)
			d.slaves[i] = ipPort
		}

		sort.Strings(d.slaves)
		log.Debugf("slaves: %v", d.slaves)
	}

	if forceUpdate {
		d.lastUpdate = time.Now()
		log.Debugf("update lastUpdate after force refresh")
	}

	for _, r := range d.LocalRedis {
		if r == d.slaves[0] {
			continue
		}
		log.Debugf("The first redis slave is owned locally, let's use it")
		if d.redisClient == nil || d.redisClient.Options().Addr != r {
			log.Debugf("Create new redis client to %s", r)
			if d.redisClient != nil {
				d.redisClient.Close()
			}
			d.redisClient = redis.NewClient(&redis.Options{
				Addr:         r,
				DialTimeout:  confTimeout,
				ReadTimeout:  confTimeout,
				WriteTimeout: confTimeout,
			}).WithContext(ctx)
		}
		return nil
	}
	log.Debugf("The first redis slave is not owned locally, let's skip for now")
	return nil
}

func (d *bucketDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	log.Debugf("Collecting (%v) ...", d)

	if err := d.updateSentinel(ctx); err != nil {
		return err
	}

	if d.redisClient == nil {
		return nil
	}

	rejects := make([]interface{}, len(confFiltersReject)+2)
	rejects[0] = d.cursor
	rejects[1] = confScanCount
	for i, r := range confFiltersReject {
		rejects[i+2] = r
	}

	info, err := scriptLua.Run(ctx, d.redisClient, confFiltersAllow, rejects).Result()
	if err != nil {
		m.IncrError("bucket", "bucket")
		log.Errorf("Error while running redis LUA script: %v", err)
		return err
	}
	type jsonUnit struct {
		Bytes   uint64 `yaml:"bytes"`
		Objects uint64 `yaml:"objects"`
		Account string `yaml:"account"`
	}
	type jsonStruct struct {
		Cursor  uint64              `yaml:"cursor"`
		Buckets map[string]jsonUnit `yaml:"buckets"`
	}
	var obj jsonStruct
	err = json.Unmarshal([]byte(info.(string)), &obj)
	if err != nil {
		log.Debugf("Go JSON: %s", info.(string))
		m.IncrError("bucket", "bucket")
		return err
	}
	d.cursor = obj.Cursor
	for bucket, data := range obj.Buckets {

		ok := true // by default allow account
		if len(d.filtersAllowAccount) > 0 {
			ok = false // by default, reject account when there are include patterns
			for _, filter := range d.filtersAllowAccount {
				if filter.Match([]byte(data.Account)) {
					ok = true // allow account
					break
				}
			}
		}
		if !ok {
			continue
		}

		for _, filter := range d.filtersRejectAccount {
			if filter.Match([]byte(data.Account)) {
				ok = false // reject account
				break
			}
		}
		if !ok {
			continue
		}

		d.sendStat("size_bytes", data.Account, bucket, data.Bytes, m)
		d.sendStat("objects", data.Account, bucket, data.Objects, m)
	}
	return nil
}

func (d *bucketDaemon) sendStat(metric string, account string, bucket string, value uint64, m *common.MetricData) {
	labels := map[string]string{
		"namespace": d.namespace,
		"account":   account,
		"bucket":    bucket,
	}
	m.Set(&common.Metric{
		Name:   "bucket_" + metric,
		Labels: labels,
		Value:  float64(value),
		TTL:    confInterval * 2,
	})
}

func (d *bucketDaemon) Close() error {
	if d.redisClient != nil {
		d.redisClient.Close()
	}
	if d.sentinelClient != nil {
		d.sentinelClient.Close()
	}
	return nil
}
