package command

import (
	"context"
	"encoding/base64"
	"errors"
	"flag"
	"fmt"
	"os"
	"os/exec"
	"strings"
	"syscall"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

var log = _log.New("command")

var confInterval = 5 * time.Second

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confInterval, "command.interval", confInterval, "[command] interval between polls")
}

type commandDaemonFactory struct {
	defaultCommands string
	commandsPath    string

	commands Commands
	version  int
}

func NewFactory(defaultCommands, commandsPath string) daemon.DaemonFactoryI {
	return &commandDaemonFactory{
		defaultCommands: defaultCommands,
		commandsPath:    commandsPath,
	}
}

func (df *commandDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	daemons := make(map[string]interface{})
	services := inventory.Services
	for _, command := range df.commands {
		_, ok := services[command.Service]
		if command.Service != "" && !ok {
			continue
		}
		daemons[command.Name] = &commandDaemonParams{
			Namespace: inventory.Namespace,
			Command:   command,
			Version:   df.version,
		}
	}

	return daemons
}

// TODO: don't do anything if the file hasn't changed
func (df *commandDaemonFactory) Reload() error {
	defaultCommands := df.defaultCommands
	commandsPath := df.commandsPath

	var commands struct {
		Commands Commands
	}

	if commandsPath == "" {
		data, err := base64.StdEncoding.DecodeString(defaultCommands)
		if err != nil {
			return err
		}

		if err := yaml.Unmarshal(data, &commands); err != nil {
			return err
		}
		log.Info("commands definition loaded from defaults")
	} else {
		f, err := os.Open(commandsPath)
		if err != nil {
			return err
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&commands); err != nil {
			return err
		}
		log.Infof("commands definition loaded from %s", commandsPath)
	}

	for _, command := range commands.Commands {
		if command.Name == "" {
			return errors.New("name is missing in one of the command definition")
		}

		if command.Command == "" {
			return fmt.Errorf("cmd is missing for command %s", command.Name)
		}

		if command.Type == "" {
			command.Type = "exec"
		}
		if command.Type != "exec" && command.Type != "shell" {
			return fmt.Errorf("wrong command type for command %s", command.Name)
		}

		if command.Interval <= 1*time.Second {
			command.Interval = confInterval
		}
		if command.Interval < confInterval {
			return fmt.Errorf(`interval can't be lower than global command.interval parameter (%v) for command "%s"`, confInterval, command.Name)
		}

		if command.Timeout <= 0 {
			command.Timeout = command.Interval
		}
		if command.Timeout > command.Interval {
			return fmt.Errorf(`timeout cannot be higher than interval (%v) for command "%s"`, command.Interval, command.Name)
		}

		log.Debugf("setup %s %v", command.Name, command.Timeout)

	}

	if !commands.Commands.Equal(df.commands) {
		log.Infoln("Commands updated")

		df.version++
	}

	df.commands = commands.Commands

	log.Infof("loaded %d commands definitions", len(df.commands))

	return nil
}

func (df *commandDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*commandDaemonParams)

	return &commandDaemon{
		command:   params.Command,
		namespace: params.Namespace,
	}
}

type Commands []*Command

type Command struct {
	Name     string        `yaml:"name"`
	Type     string        `yaml:"type"`
	Command  string        `yaml:"command"`
	Interval time.Duration `yaml:"interval"`
	Timeout  time.Duration `yaml:"timeout"`
	Service  string        `yaml:"service"`
	LastLine bool          `yaml:"last_line"`
}

func (commands1 Commands) Equal(commands2 Commands) bool {
	if len(commands1) != len(commands2) {
		return false
	}

	for _, v1 := range commands1 {
		found := false
		for _, v2 := range commands2 {
			if v1.Name != v2.Name {
				continue
			}

			if !v1.Equal(v2) {
				return false
			}

			found = true
		}

		if !found {
			return false
		}
	}

	return true
}

// To exclude Regex
func (v1 *Command) Equal(v2 *Command) bool {
	return v2 != nil &&
		v1.Name == v2.Name &&
		v1.Type == v2.Type &&
		v1.Command == v2.Command &&
		v1.Interval == v2.Interval &&
		v1.Timeout == v2.Timeout
}

type commandDaemonParams struct {
	Namespace string
	Command   *Command
	Version   int
}

type commandDaemon struct {
	daemon.BaseDaemon

	namespace string
	command   *Command
}

func (d *commandDaemon) Interval() time.Duration {
	if d.command.Interval > 0 {
		return d.command.Interval
	}
	return confInterval
}

func (d *commandDaemon) Delay() time.Duration {
	return 0
}

func (d *commandDaemon) Collect(ctx context.Context, m *common.MetricData) error {

	ret, lastLine, err := d.runCommand(ctx)
	// TODO: do something if err == exec.ErrNotFound
	if err != nil {
		m.IncrError("command", d.command.Name)
		log.Warnln("error while running command", d.command.Command, d.command.Name, err)
		return nil
	}

	log.Debugf(`%s returned: %d: "%s"`, d.command.Name, ret, lastLine)

	labels := map[string]string{
		"namespace": d.namespace,
		"command":   d.command.Name,
		"interval":  fmt.Sprintf("%d", int(d.command.Interval.Seconds())),
	}

	if d.command.LastLine {
		labels["last_line"] = lastLine
	}

	m.Set(&common.Metric{
		Name:   "command",
		Labels: labels,
		Value:  float64(ret),
	})

	return nil
}

func (d *commandDaemon) runCommand(ctx context.Context) (int, string, error) {
	var out []byte
	var err error
	var ret int = 0
	var timeout bool = false

	log.Debugf(`%s: %s "%s"`, d.command.Name, d.command.Type, d.command.Command)
	if d.command.Type == "shell" {
		out, err = common.ExecArray(ctx, d.command.Timeout, "/bin/sh", "-c", d.command.Command)
	} else {
		out, err = common.ExecString(ctx, d.command.Timeout, d.command.Command)
	}

	if err != nil {

		if errors.Is(err, context.DeadlineExceeded) {
			timeout = true
			goto exit
		}

		if exitError, ok := err.(*exec.ExitError); ok {
			if status, ok := exitError.Sys().(syscall.WaitStatus); ok {
				if status.Signaled() && status.Signal() == syscall.SIGKILL {
					timeout = true
					goto exit
				}

				if status.Exited() {
					ret = status.ExitStatus()
					goto exit
				}
			}
		}

		return 0, "", err
	}

exit:

	if timeout {
		// from /usr/include/asm/errno.h
		// #define ETIMEDOUT       110
		return 110, "command timeout", nil
	}

	if !d.command.LastLine {
		return ret, "", nil
	}

	outFmt := strings.Split(strings.Trim(strings.TrimSuffix(string(out), "\n"), " "), "\n")

	return ret, outFmt[len(outFmt)-1], nil
}
