package openio

import (
	"flag"
	"fmt"
	"sort"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "openio.timeout", confTimeout, "[openio] http timeout")
	fs.DurationVar(&confInterval, "openio.interval", confInterval, "[openio] interval between polls")
}

var Factory = &openioDaemonFactory{}

type openioDaemonFactory struct{}

func (df *openioDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})

	proxyServices := inventory.Services["oioproxy"]
	if len(proxyServices) == 0 {
		return params
	}
	proxyAddrs := make([]string, len(proxyServices))
	for i, service := range proxyServices {
		proxyAddrs[i] = fmt.Sprintf("%s:%d", service.IP, service.Port)
	}
	// Sorting because a different order should not cause a reload
	sort.Strings(proxyAddrs)

	df.scoreDaemons(inventory, proxyAddrs, params)
	df.volumeDaemons(inventory, params)
	df.metaDaemons(inventory, proxyAddrs, params)
	df.rawxDaemons(inventory, params)

	return params
}

func (df *openioDaemonFactory) scoreDaemons(inventory *common.Inventory, proxyAddrs []string, params map[string]interface{}) {
	for serviceType, services := range inventory.Services {
		if _, ok := SCORE_SERVICE_TYPES[serviceType]; !ok {
			continue
		}

		scoreServices := make([]*scoreService, len(services))
		for i, service := range services {
			scoreServices[i] = &scoreService{
				ID:     service.ID,
				IP:     service.IP,
				Port:   service.Port,
				Volume: service.Volume,
			}
		}

		// Sorting because a different order should not cause a reload
		sort.Slice(scoreServices, func(i, j int) bool {
			return scoreServices[i].ID > scoreServices[j].ID
		})

		daemonHash := "score:" + serviceType
		params[daemonHash] = &openioScoreDaemonParams{
			Namespace:     inventory.Namespace,
			ServiceType:   serviceType,
			ScoreServices: scoreServices,
			ProxyAddrs:    proxyAddrs,
		}
	}
}

func (df *openioDaemonFactory) volumeDaemons(inventory *common.Inventory, params map[string]interface{}) {
	for serviceType, services := range inventory.Services {
		for _, service := range services {
			if service.Volume == "" {
				continue
			}

			daemonHash := "volume:" + service.ID
			params[daemonHash] = &openioVolumeDaemonParams{
				Namespace:     inventory.Namespace,
				ServiceType:   serviceType,
				ServiceID:     service.ID,
				ServiceIP:     service.IP,
				ServicePort:   service.Port,
				ServiceVolume: service.Volume,
			}
		}
	}
}

func (df *openioDaemonFactory) metaDaemons(inventory *common.Inventory, proxyAddrs []string, params map[string]interface{}) {
	for serviceType, services := range inventory.Services {
		if !strings.HasPrefix(serviceType, "meta") {
			continue
		}

		for _, service := range services {
			daemonHash := "meta:" + service.ID
			params[daemonHash] = &openioMetaDaemonParams{
				Namespace:     inventory.Namespace,
				ServiceType:   serviceType,
				ServiceID:     service.ID,
				ServiceIP:     service.IP,
				ServicePort:   service.Port,
				ServiceVolume: service.Volume,
				ProxyAddrs:    proxyAddrs,
			}
		}
	}
}

func (df *openioDaemonFactory) rawxDaemons(inventory *common.Inventory, params map[string]interface{}) {
	for _, service := range inventory.Services["rawx"] {
		daemonHash := "rawx:" + service.ID
		params[daemonHash] = &openioRawxDaemonParams{
			Namespace:     inventory.Namespace,
			ServiceID:     service.ID,
			ServiceIP:     service.IP,
			ServicePort:   service.Port,
			ServiceVolume: service.Volume,
		}
	}
}

func (df *openioDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	switch params := paramsI.(type) {
	case *openioScoreDaemonParams:
		return &openioScoreDaemon{
			namespace:     params.Namespace,
			serviceType:   params.ServiceType,
			scoreServices: params.ScoreServices,
			proxyAddrs:    params.ProxyAddrs,
		}
	case *openioVolumeDaemonParams:
		return &openioVolumeDaemon{
			namespace:     params.Namespace,
			serviceType:   params.ServiceType,
			serviceID:     params.ServiceID,
			serviceIP:     params.ServiceIP,
			servicePort:   params.ServicePort,
			serviceVolume: params.ServiceVolume,
		}
	case *openioMetaDaemonParams:
		return &openioMetaDaemon{
			namespace:     params.Namespace,
			serviceType:   params.ServiceType,
			serviceID:     params.ServiceID,
			serviceIP:     params.ServiceIP,
			servicePort:   params.ServicePort,
			serviceVolume: params.ServiceVolume,
			proxyAddrs:    params.ProxyAddrs,
		}
	case *openioRawxDaemonParams:
		return &openioRawxDaemon{
			namespace:     params.Namespace,
			serviceID:     params.ServiceID,
			serviceIP:     params.ServiceIP,
			servicePort:   params.ServicePort,
			serviceVolume: params.ServiceVolume,
		}
	}

	return nil
}
